'******************************************************************************
'*  io.spin - handle input and output requests sent by pm80
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: io.spin,v 1.20 2010-03-25 09:40:35 pm Exp $
'******************************************************************************
CON

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break
	io_cmd_kbd		=	$04		' I/O emulator command read TRS-80 keyboard

	baud			=	115200		' baud rate

VAR
	long	cog
	long	io_command
	long	params[2]
	long	rc
	word	topy
	word	boty
	word	row
	word	y0
	byte	curx
	byte	cury
	byte	de[32]					' directory entry
	byte	name[12+1]
	byte	datetime[20+1]
	byte	buff[32]

OBJ
	fat	: "fatfs"
	video	: "trs80_video"
	kbd	: "Keyboard_rr002"
#ifdef TriBladeProp
	xmm	: "TriBladeProp_Blade2_204"		' Extended memory manager
#endif
#ifdef RamBladeProp
	xmm	: "RamBlade_203_002"			' Extended memory manager
#endif
#ifdef DracBladeProp
	xmm	: "Dracblade"				' Extended memory manager
#endif

PUB start(SI, SO, doPin, clkPin, diPin, csPin, kbDpin, kbCpin, auPin, tvPin, screen, iobuff) : okay | i, dent, sec
	stop

	screen_ptr := screen
	iobuff_ptr := iobuff
	chr(12)
	str(string("********* TRS-80 emulator on the Propeller *********",13,10))

	okay := \fat.start(iobuff_ptr, doPin, clkPin, diPin, csPin)
	if okay < 0
		failure(string("FAT + sdspi"),okay)
		return -1
	success(string("FAT + sdspi"))

	params[0] := screen				' 1K video RAM
	params[1] := tvPin				' video pin
	video.start(@params)

	okay := \kbd.startx(kbDpin, kbCpin, %100, 40)
	if okay < 0
		failure(string("Keyboard"),okay)
		return
	success(string("Keyboard"))
	kbd_states := kbd.states_ptr
	kbd_change := kbd.change_ptr

	dira_mask := 1 << auPin
	ctra_mode := %00110 << 26 | auPin		' set to single ended DUTY mode

	okay := cog := COGNEW(@entry, @io_command) + 1
	if okay =< 0
		failure(string("I/O"),okay)
		return -1
	success(string("I/O"))

	okay := \xmm.start
	if okay < 0
		failure(string("XMM"), okay)
		return -1
	success(string("XMM"))
	xmm_command := xmm.rendezvous
	xmm_hubaddr := xmm_command + 1*4
	xmm_ramaddr := xmm_command + 2*4
	xmm_blocklen := xmm_command + 3*4
	xmm_errx := xmm_command + 4*4

	' clear the Z80 RAM
	longfill(iobuff_ptr, 0, 512/4)
	repeat i from $4000 to $fe00 step $200
		xmm.DoCmd("W", iobuff_ptr, i, $200)

	okay := \fat.mount
	if okay < 0
		failure(string("FAT mount"), okay)
		return -1
	str(string("FAT"))
	dec(fat.get_type)
	str(string(" directory entries: "))
	dec(fat.get_dir_size)
	crlf

	dent := open_file(string("TRS80.ROM"))
	if dent < 0
		return -1
	sec := fat.get_dent_first_sector(@de)
	repeat i from $0000 to $2e00 step $200
		fat.read_sector(sec)
		xmm.DoCmd("W", iobuff_ptr, i, $200)
		sec := fat.get_dent_next_sector(@de, sec)

	' Patch the input number at MEM SIZE? to not wait for input
	bytefill(iobuff_ptr, 0, 3)
	xmm.DoCmd("W", iobuff_ptr, $00bb, 3)

	row := 0
	topy := 0
	str(string("Select file to load to memory: <UP> <DN> <RETURN>"))
	repeat
		boty := 9999
		repeat y0 from 0 to 6
			rc := direntry(topy + y0, @buff)
			if rc < 0
				bytefill(@buff, " ", 14)
				buff[14] := 0
				if boty == 9999
					boty := y0
			gotoxy(0, 8+y0)
			if y0 == row
				str(string("==> "))
			else
				str(string("    "))
			str(@buff)
			if y0 == row
				str(string(" <==          "))
			else
				str(string("              "))
		case kbd.getkey
			$c3:	' down arrow
				if row < 6
					if row < boty
						row++
				else
					if boty == 9999
						topy++
			$c2:	' up arrow
				if row > 0
					row--
				elseif topy > 0
					topy--
			$0d:
				bytefill(@buff, " ", 32)
				buff[31] := 0
				repeat y0 from 0 to 6
					gotoxy(0, 8+y0)
					str(@buff)
				gotoxy(0, 8)
				rc := direntry(topy + row, @buff)
				quit
		
	cmd_load(@buff)
	

	str(string("Press <RETURN> to start...",13,10))
	repeat until kbd.getkey == 13

	return 0

PUB stop
'' Stop CPU simulator - frees a cog
	if cog
		cogstop(cog~ - 1)

PRI gotoxy(x,y)
	curx := (x #> 0) <# 63
	cury := (y #> 0) <# 15

PRI chr(ch)
	case ch
		13:
			curx := 0
		12:
			longfill(screen_ptr, $20202020, 64*16/4)
			curx := 0
			cury := 0
		10:
			cury += 1
			if cury > 15
				longmove(screen_ptr, screen_ptr + 64, 64*15/4)
				longfill(screen_ptr + 64*15, $20202020, 64/4)
				cury := 15
		9:
			repeat until (curx & 7) == 0
				chr(32)
		8:
			if curx > 0
				--curx
		other:
			byte[screen_ptr][cury << 6 + curx] := ch
			curx++
			if curx == 64
				chr(13)
				chr(10)

PRI str(str_ptr) | i
	repeat i from 0 to strsize(str_ptr) - 1
		chr(byte[str_ptr][i])

PRI crlf
	chr(13)
	chr(10)

PRI dec_size(value)
	if value > 1048576
		dec(value/1048576)
		chr(".")
		dec(10*(value // 1048576) / 1048576)
		str(string("MB"))
	elseif value > 1024
		dec(value/1024)
		chr(".")
		dec(10*(value // 1024) / 1024)
		str(string("KB"))
	else
		dec(value)

PRI dec(value) | _i
'' Print a decimal number
	result := 0
	if value < 0
		chr("-")
		if (value == NEGX)
			chr("2")
			value += 2_000_000_000
		-value
	_i := 1_000_000_000
	repeat 10
		if value => _i
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			chr("0")
		_i /= 10

PRI dech(value) | _i
'' Print a decimal number with 1000s tick marks

	if value < 0
		chr("-")
		if (value == NEGX)
			chr("2")
			value += 2_000_000_000
		-value

	_i := 1_000_000_000
	repeat 10
		if value => _i
			if _i == 100_000_000 or _i == 100_000 or _i == 100
				chr("'")
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			if _i == 100_000_000 or _i == 100_000 or _i == 100
				chr("'")
			chr("0")
		_i /= 10

PRI hex(value, digits)
'' Print a hexadecimal number

	value <<= (8 - digits) << 2
	repeat digits
		chr(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PRI bin(value, digits)
'' Print a binary number

	value <<= 32 - digits
	repeat digits
		chr((value <-= 1) & 1 + "0")

PRI failure(who, val)
	str(who)
	str(string(" failed: "))
	dec(val)
	crlf

PRI success(who) | i, n, n1, n2, n3, n4, n5, n6, n7
	str(who)
	str(string(" initialized, "))
	dec(free_cogs)
	str(string(" cogs free.", 13, 10))

PUB free_cogs : free | i, cogs[8], jmp_0
	jmp_0 := %010111_0001_1111_000000000_000000000
	repeat while (cogs[free] := cognew(@jmp_0, 0)) => 0 
		free++
	if free
		repeat i from 0 to free - 1
			cogstop(cogs[i])
	return free

PRI direntry(n, buf) | rec, size, i
	rec := fat.opendir(@de)
	repeat while rec => 0
		if byte[@de][fat#dent_name] <> $e5	' deleted entry?
			byte[@de][fat#dent_attr] := 0
			if strcomp(@de + fat#dent_ext, string("CMD"))
				if n == 0
					fat.get_dent_name(@de, @name)
					i := strsize(@name)
					bytemove(buf, @name, i)
					byte[buf][i] := 0
					return 0
				n--
			rec := fat.readdir(rec, @de)
	return -1

PRI open_file(fname_ptr) : dent | size, sec
	dent := fat.find_file(fname_ptr)
	if dent < 0
		str(fname_ptr)
		str(string(" not found.",13,10))
		return -1
	bytemove(@de, dent, 32)
	fat.get_dent_name(dent, @name)
	size := fat.get_dent_size(@de)
	sec := fat.get_dent_first_sector(@de)
	fat.get_dent_date_time(@de, @datetime)
	str(@name)
	str(string(9,"sector "))
	dec(sec)
	str(string(9,"size "))
	dec_size(size)
	crlf

PRI cmd_load(fname_ptr) | dent, sec, nsec, pos, state, badr, blen, bmin, bmax
	dent := open_file(fname_ptr)
	if dent < 0
		return -1
	fat.get_dent_name(dent, @name)
	sec := fat.get_dent_first_sector(@de)

	pos := $200
	state := 0
	bmin := $ffff
	bmax := $0000
	repeat
		if pos == $200
			if sec < 0
				quit
			nsec := fat.get_dent_next_sector(@de, sec)
			fat.read_sector(sec)
			sec := nsec
			pos := 0
		case state
			0:	' expecting a header
				case byte[iobuff_ptr][pos++]
					$01:	' CMD load block header
						state := 1
					$02:	' CMD entry point header
						state := 7
					other:	' CMD comment block
						state := 5

			1:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				state := 2

			2:	' expecting address LSB
				blen--
				badr := byte[iobuff_ptr][pos++]
				state := 3

			3:	' expecting address MSB
				blen--
				badr += 256 * byte[iobuff_ptr][pos++]
				state := 4
				if blen =< 0
					blen += 256
				bmin := bmin <# badr

			4:	' expecting blen bytes of data
				xmm.DoCmd("W", iobuff_ptr + pos, badr, 1)
				badr++
				pos++
				blen--
				if blen == 0
					bmax := bmax #> badr
					state := 0

			5:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				if blen == 0
					blen := 256
				state := 6

			6:	' ignoring comment data
				pos++
				blen--
				if blen == 0
					state := 0

			7:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				state := 8

			8:	' expecting entry point LSB
				blen--
				badr := byte[iobuff_ptr][pos++]
				state := 9

			9:	' expecting entry point MSB
				blen--
				badr += 256 * byte[iobuff_ptr][pos++]
				str(string("Loaded from "))
				hex(bmin, 4)
				str(string(".."))
				hex(bmax - 1, 4)
				crlf
				str(string("Entry point "))
				hex(badr, 4)
				crlf
				xmm.DoCmd("W", @badr, $0103, 2)
				quit

PUB command_addr
'' Return the address of the command long
	return @io_command

DAT
'---------------------------------------------------------------------------------------------------------

		org	0

'---------------------------------------------------------------------------------------------------------
entry
iocmd_ptr	mov	iocmd_ptr, PAR
		jmp	#run

screen_ptr	long	0-0
addr		long	0
cur_x		long	0
cur_y		long	0
iobuff_ptr	long	0-0
x20202020	long	$20202020
x10_000		long	10_000

dira_mask	long	0-0
ctra_mode	long	0-0
frqa_0		long	$7850_0000				' 0.46V of 1V, ie. 46%
frqa_1		long	$0000_0000				' 0V
frqa_2		long	$d999_0000				' 0.85V of 1V, ie. 85%
frqa_3		long	$0000_0000				' 0V

xmm_command	long	0-0
xmm_hubaddr	long	0-0
xmm_ramaddr	long	0-0
xmm_blocklen	long	0-0
xmm_errx	long	0-0

kbd_lookup	long	@@@kbd_lookup_table
kbd_rows	long	0, 0, 0, 0, 0, 0, 0, 0
kbd_states	long	0-0
kbd_change	long	0-0
kbd_index	long	0
kbd_shift	long	0
kbd_data	long	0

minus1		long	$ffffffff
low_24bits	long	$00ffffff
low_16bits	long	$0000ffff
msg_break_ptr	long	@@@msg_break

iocmd		long	0
port		long	0
data		long	0
null		long	0

t1		long	0
t2		long	0
t3		long	0
t4		long	0

strptr		long	0
hexval		long	0
hexcnt		long	0

run
		wrlong	null, iocmd_ptr

		mov	DIRA, dira_mask			' enable the sound output
		mov	CTRA, ctra_mode			' set the counter mode to single ended DUTY mode
		mov	FRQA, frqa_0			' set the default output pulse width

command
		rdlong	t1, kbd_change		WZ
	if_nz	jmp	#keyboard
		rdlong	iocmd, iocmd_ptr
		mov	t1, iocmd
		shr	t1, #24			WZ
	if_z	jmp	#command
		cmp	t1, #io_cmd_out		WZ
	if_z	jmp	#output
		cmp	t1, #io_cmd_in		WZ
	if_z	jmp	#input
		cmp	t1, #io_cmd_break	WZ
	if_z	jmp	#break
		cmp	t1, #io_cmd_kbd		WZ
	if_z	jmp	#key_matrix
command_0
		and	iocmd, low_24bits
		wrlong	iocmd, iocmd_ptr
		jmp	#command

'*******************************************
'
' Read i/o port and data to write to
'
output
		mov	port, iocmd			' get command word
		shr	port, #8			' extract port
		and	port, #$ff			' lower 8 bits
		mov	data, iocmd			' get command word
		and	data, #$ff			' extract data

		cmp	port, #$ff		WZ	' port = $FF?
	if_z	call	#output_ff

		jmp	#command_0


output_ff
		mov	t1, data
		and	t1, #3
		add	t1, #frqa_0
		movs	:padr, t1
		mov	t1, data
:padr		mov	t2, 0-0				' get audio level
		mov	FRQA, t2			' write to FRQA
		' TODO: check the other bits of data
output_ff_ret
		ret

'*******************************************
'
' Read i/o port to read from
'
input
		andn	iocmd, #$ff			' clear the input data
		mov	port, iocmd			' get command word
		shr	port, #8
		and	port, #$ff

		' TODO: port FF

		or	iocmd, #$ff			' return $ff for other ports
		jmp	#command_0

'*******************************************
'
' Output a character to the screen
'
conout_chr
		mov	addr, cur_y			' y
		shl	addr, #6			' * 64
		add	addr, cur_x			' + x
		add	addr, screen_ptr		' + screen address

		cmp	data, #8		WZ	' backspace?
	if_z	jmp	#:bsp
		cmp	data, #9		WZ	' tab?
	if_z	jmp	#:tab
		cmp	data, #10		WZ	' line feed?
	if_z	jmp	#:lf
		cmp	data, #12		WZ	' clear screen?
	if_z	jmp	#:cls
		cmp	data, #13		WZ	' carriage return?
	if_z	jmp	#:cr
:out
		wrbyte	data, addr
		add	cur_x, #1
:wrap
		cmp	cur_x, #64		WZ
	if_nz	jmp	#:exit
		mov	cur_x, #0
:lf
		add	cur_y, #1
		cmp	cur_y, #16		WZ
	if_nz	jmp	#:exit
		call	#scroll
		mov	cur_y, #15
		jmp	#:exit

:bsp
		cmp	cur_x, #0		WZ, WC
	if_be	jmp	#:exit
		sub	cur_x, #1
		jmp	#:exit
:tab
		mov	data, #32
		wrbyte	data, addr
		add	addr, #1
		add	cur_x, #1
		test	cur_x, #7		WZ
	if_nz	jmp	#:tab
		jmp	#:wrap
:cr
		mov	cur_x, #0
		jmp	#:exit
:cls
		mov	addr, screen_ptr
		mov	cur_y, #64*16/4
:loop
		wrlong	x20202020, addr
		add	addr, #4
		djnz	cur_y, #:loop
		mov	cur_x, #0
:exit
conout_chr_ret
		ret		

'*******************************************
'
' Scroll the screen one line up
'
scroll
		mov	addr, screen_ptr
		mov	cur_y, #15
:row
		mov	cur_x, #64
:col
		add	addr, #64
		rdlong	data, addr
		sub	addr, #64
		wrlong	data, addr
		add	addr, #4
		djnz	cur_x, #:col
		djnz	cur_y, #:row
		mov	cur_x, #64/4
:fill
		wrlong	x20202020, addr
		add	addr, #4
		djnz	cur_x, #:fill
scroll_ret
		ret


'*******************************************
'
' Output a byte in hexval to the console
'
conout_hex_b
		mov	hexcnt, #2
		shl	hexval, #24
		jmp	#conout_hex

'*******************************************
'
' Output a word in hexval to the console
'
conout_hex_w
		mov	hexcnt, #4
		shl	hexval, #16
		jmp	#conout_hex

'*******************************************
'
' Output a long in hexval to the console
'
conout_hex_l
		mov	hexcnt, #8
conout_hex
		mov	data, hexval
		shr	data, #28
		cmp	data, #10	WZ, WC
	if_b	add	data, #48
	if_ae	add	data, #55
		call	#conout_chr
		shl	hexval, #4
		djnz	hexcnt, #conout_hex
conout_hex_ret
conout_hex_l_ret
conout_hex_w_ret
conout_hex_b_ret
		ret

'*******************************************
'
' Output a carriage return to the console
'
conout_space
		mov	data, #32
		call	#conout_chr
conout_space_ret
		ret

'*******************************************
'
' Output a carriage return to the console
'
conout_crlf
		mov	data, #13
		call	#conout_chr
		mov	data, #10
		call	#conout_chr
conout_crlf_ret
		ret

'*******************************************
'
' Output a string at strptr to the console
'
conout_string
		rdbyte	data, strptr	WZ
		add	strptr, #1
	if_z	jmp	#conout_string_ret
		call	#conout_chr
		jmp	#conout_string
conout_string_ret
		ret

'*******************************************
'
' Break - issued when the Z80 encounters
' a HALT instruction.
'
break
		mov	t1, iocmd			' read the address of the register file
		shr	t1, #8				' bits 23..8 to 15..0
		and	t1, low_16bits			' mask only bits 15..0
		mov	strptr, msg_break_ptr	

		mov	t2, #8
:loop_r16	call	#conout_string
		rdword	hexval, t1			' read BC, DE, HL, AF, BC', DE', HL', AF'
		call	#conout_hex_w
		add	t1, #2
		djnz	t2, #:loop_r16

		call	#conout_string			' new line and print "R:"
		rdbyte	hexval, t1			' R
		add	t1, #1
		rdbyte	t2, t1				' R2
		and	hexval, #%01111111		' R register's bit #7
		or	hexval, t2			' does never change from the last written value
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdbyte	hexval, t1			' IFF1+2/IM
		mov	t2, t1				' save value
		and	hexval, #1			' IFF1 (interrupt flip-flop 1)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #1			' shift bit #1 to #0
		and	hexval, #1			' IFF2 (interrupt flip-flop 2)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #2			' shift bits #3..2 to #1..0
		and	hexval, #3			' IM (interrupt mode)
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdbyte	hexval, t1			' I (interrupt vector register)
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdword	hexval, t1			' IX
		call	#conout_hex_w
		add	t1, #2

		call	#conout_string
		rdword	hexval, t1			' IY
		call	#conout_hex_w
		add	t1, #2

		call	#conout_string
		rdword	hexval, t1			' SP
		call	#conout_hex_w
		add	t1, #2

		call	#conout_string
		rdword	hexval, t1			' PC
		mov	t3, hexval
		call	#conout_hex_w
		call	#conout_string
		add	t1, #2				' skip to opcode

		wrlong	t3, xmm_ramaddr			' as XMM ram address
		mov	t2, #4				' 4 bytes
		wrlong	t2, xmm_blocklen		' blocklen to read
		mov	t2, #"R"			' XMM read command
		wrlong	t2, xmm_command
:wait		rdlong	t2, xmm_command		WZ
	if_nz	jmp	#:wait

		rdbyte	hexval, t1
		call	#conout_hex_b

		call	#conout_crlf

		jmp	#command_0

'*******************************************
'
' Read a keyboard key and translate it
'
keyboard
		wrlong	null, kbd_change		' reset keyboard changed flag

		mov	t1, #kbd_rows
		mov	t3, #8
:clear		movd	:row_clr, t1
		add	t1, #1
:row_clr	mov	0-0, #0
		djnz	t3, #:clear

		mov	t1, kbd_lookup			' pointer to the lookup table
		mov	t2, kbd_states			' pointer to the kbd_states table
		mov	t3, #256			' 256 key states to test
		mov	data, #1			' test bitmask
		rdlong	kbd_data, t2			' get first long (32 bits) from kbd_states
		add	t2, #4
:loop
		rdbyte	kbd_shift, t1			' get the TRS80 row, column and shift index
		add	t1, #1				' next entry
		cmp	kbd_shift, #$ff		WZ	' key is to be ignored?
	if_z	jmp	#:ignore
		mov	kbd_index, kbd_shift
		shr	kbd_index, #4			' extract TRS80 keyboard row 0..7
		and	kbd_index, #%111		' extract kbd column 0..7
		add	kbd_index, #kbd_rows		' index to kbd_rows
		movd	:rows_bit, kbd_index
		and	kbd_shift, #%111		' extract kbd column 0..7
		mov	t4, #1
		rol	t4, kbd_shift			' rotate bitmask
		test	kbd_data, data		WZ	' if state bit is set, set non-zero
:rows_bit
	if_nz	or	0-0, t4				' set key bit on non-zero
:ignore
		rol	data, #1		WC	' next bit to test
	if_c	rdlong	kbd_data, t2			' fetch next long from keyboard states
	if_c	add	t2, #4				' advance pointer to next kbd_states entry
		djnz	t3, #:loop
		jmp	#command

'*******************************************
'
' Read the TRS-80 keyboard matrix
' Address is in io_port, address bits 0..7
' determine the keyboard row(s) to read
'
key_matrix
		mov	port, iocmd			' get command word
		shr	port, #8
		and	port, low_16bits
		mov	t1, #0				' resulting bits
		mov	t2, #kbd_rows			' address of the keyboard matrix
		mov	t3, #8				' 8 of them
		mov	data, #$01			' bit mask for the address lines
:loop		movs	:get_row, t2
		add	t2, #1				' next keyboard row entry
		test	port, data		WZ	' test port address A0, A1, A2 ...
:get_row	mov	kbd_data, 0-0			' get the keyboard row value
	if_nz	or	t1, kbd_data			' if address bit was set, OR the data
		rol	data, #1			' rotate mask left
		djnz	t3, #:loop			' 8 times
		or	iocmd, t1			' return it in the command word bits 7..0
		jmp	#command_0

		fit	$1f0

'***************************************************************************
'*   w/o SHIFT                             with SHIFT
'*   +-------------------------------+     +-------------------------------+
'*   | 0   1   2   3   4   5   6   7 |     | 0   1   2   3   4   5   6   7 |
'*+--+---+---+---+---+---+---+---+---+  +--+---+---+---+---+---+---+---+---+
'*|0 | @ | A | B | C | D | E | F | G |  |0 | ` | a | b | c | d | e | f | g |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|1 | H | I | J | K | L | M | N | O |  |1 | h | i | j | k | l | m | n | o |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|2 | P | Q | R | S | T | U | V | W |  |2 | p | q | r | s | t | u | v | w |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|3 | X | Y | Z | [ | \ | ] | ^ | _ |  |3 | x | y | z | { | | | } | ~ |   |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|4 | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 |  |4 |   | ! | " | # | $ | % | & | ' |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|5 | 8 | 9 | : | ; | , | - | . | / |  |5 | ( | ) | * | + | < | = | > | ? |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|6 |ENT|CLR|BRK|UP |DN |LFT|RGT|SPC|  |6 |ENT|CLR|BRK|UP |DN |LFT|RGT|SPC|
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|7 |SHF|ALT|PUP|PDN|INS|DEL|CTL|END|  |7 |SHF|ALT|PUP|PDN|INS|DEL|CTL|END|
'*+--+---+---+---+---+---+---+---+---+  +--+---+---+---+---+---+---+---+---+
'***************************************************************************

' format rrrrcccs - rrrr = row number, ccc = column number, s = shifted
kbd_lookup_table
	byte	$ff			''	00
	byte	$ff			''	01
	byte	$ff			''	02
	byte	$ff			''	03
	byte	$ff			''	04
	byte	$ff			''	05
	byte	$ff			''	06
	byte	$ff			''	07
	byte	$ff			''	08
	byte	6<<4|6<<1|0		''	09	Tab
	byte	$ff			''	0A
	byte	$ff			''	0B
	byte	$ff			''	0C
	byte	6<<4|0<<1|0		''	0D	Enter
	byte	$ff			''	0E
	byte	$ff			''	0F
	byte	$ff			''	10
	byte	$ff			''	11
	byte	$ff			''	12
	byte	$ff			''	13
	byte	$ff			''	14
	byte	$ff			''	15
	byte	$ff			''	16
	byte	$ff			''	17
	byte	$ff			''	18
	byte	$ff			''	19
	byte	$ff			''	1A
	byte	$ff			''	1B
	byte	$ff			''	1C
	byte	$ff			''	1D
	byte	$ff			''	1E
	byte	$ff			''	1F
	byte	6<<4|7			''	20	Space
	byte	4<<4|1			''	21	!
	byte	4<<4|2			''	22	"
	byte	4<<4|3			''	23	#
	byte	4<<4|4			''	24	$
	byte	4<<4|5			''	25	%
	byte	4<<4|6			''	26	&
	byte	4<<4|7			''	27	'
	byte	5<<4|0			''	28	(
	byte	5<<4|1			''	29	)
	byte	5<<4|2			''	2A	*
	byte	5<<4|3			''	2B	+
	byte	5<<4|4			''	2C	,
	byte	5<<4|5			''	2D	-
	byte	5<<4|6			''	2E	.
	byte	5<<4|7			''	2F	/
	byte	4<<4|0			''	30	0
	byte	4<<4|1			''	31	1
	byte	4<<4|2			''	32	2
	byte	4<<4|3			''	33	3
	byte	4<<4|4			''	34	4
	byte	4<<4|5			''	35	5
	byte	4<<4|6			''	36	6
	byte	4<<4|7			''	37	7
	byte	5<<4|0			''	38	8
	byte	5<<4|1			''	39	9
	byte	5<<4|2			''	3A	:
	byte	5<<4|3			''	3B	;
	byte	5<<4|4			''	3C	<
	byte	5<<4|5			''	3D	=
	byte	5<<4|6			''	3E	>
	byte	5<<4|7			''	3F	?
	byte	0<<4|0			''	40	@	
	byte	0<<4|1			''	41	A
	byte	0<<4|2			''	42	B
	byte	0<<4|3			''	43	C
	byte	0<<4|4			''	44	D
	byte	0<<4|5			''	45	E
	byte	0<<4|6			''	46	F
	byte	0<<4|7			''	47	G
	byte	1<<4|0			''	48	H
	byte	1<<4|1			''	49	I
	byte	1<<4|2			''	4A	J
	byte	1<<4|3			''	4B	K
	byte	1<<4|4			''	4C	L
	byte	1<<4|5			''	4D	M
	byte	1<<4|6			''	4E	N
	byte	1<<4|7			''	4F	O
	byte	2<<4|0			''	50	P
	byte	2<<4|1			''	51	Q
	byte	2<<4|2			''	52	R
	byte	2<<4|3			''	53	S
	byte	2<<4|4			''	54	T
	byte	2<<4|5			''	55	U
	byte	2<<4|6			''	56	V
	byte	2<<4|7			''	57	W
	byte	3<<4|0			''	58	X
	byte	3<<4|1			''	59	Y
	byte	3<<4|2			''	5A	Z
	byte	3<<4|3			''	5B	[
	byte	3<<4|4			''	5C	\
	byte	3<<4|5			''	5D	]
	byte	3<<4|6			''	5E	^
	byte	3<<4|7			''	5F	_
	byte	0<<4|0			''	60	`	
	byte	0<<4|1			''	61	a
	byte	0<<4|2			''	62	b
	byte	0<<4|3			''	63	c
	byte	0<<4|4			''	64	d
	byte	0<<4|5			''	65	e
	byte	0<<4|6			''	66	f
	byte	0<<4|7			''	67	g
	byte	1<<4|0			''	68	h
	byte	1<<4|1			''	69	i
	byte	1<<4|2			''	6A	j
	byte	1<<4|3			''	6B	k
	byte	1<<4|4			''	6C	l
	byte	1<<4|5			''	6D	m
	byte	1<<4|6			''	6E	n
	byte	1<<4|7			''	6F	o
	byte	2<<4|0			''	70	p
	byte	2<<4|1			''	71	q
	byte	2<<4|2			''	72	r
	byte	2<<4|3			''	73	s
	byte	2<<4|4			''	74	t
	byte	2<<4|5			''	75	u
	byte	2<<4|6			''	76	v
	byte	2<<4|7			''	77	w
	byte	3<<4|0			''	78	x
	byte	3<<4|1			''	79	y
	byte	3<<4|2			''	7A	z
	byte	3<<4|3			''	7B	{
	byte	3<<4|4			''	7C	|
	byte	3<<4|5			''	7D	}
	byte	3<<4|6			''	7E	~
	byte	3<<4|7			''	7F	[]

	byte	$ff			''	80-BF	(future international character support)
	byte	$ff			''	81
	byte	$ff			''	82
	byte	$ff			''	83
	byte	$ff			''	84
	byte	$ff			''	85
	byte	$ff			''	86
	byte	$ff			''	87
	byte	$ff			''	88
	byte	$ff			''	89
	byte	$ff			''	8A
	byte	$ff			''	8B
	byte	$ff			''	8C
	byte	$ff			''	8D
	byte	$ff			''	8E
	byte	$ff			''	8F
	byte	$ff			''	90
	byte	$ff			''	91
	byte	$ff			''	92
	byte	$ff			''	93
	byte	$ff			''	94
	byte	$ff			''	95
	byte	$ff			''	96
	byte	$ff			''	97
	byte	$ff			''	98
	byte	$ff			''	99
	byte	$ff			''	9A
	byte	$ff			''	9B
	byte	$ff			''	9C
	byte	$ff			''	9D
	byte	$ff			''	9E
	byte	$ff			''	9F
	byte	$ff			''	A0
	byte	$ff			''	A1
	byte	$ff			''	A2
	byte	$ff			''	A3
	byte	$ff			''	A4
	byte	$ff			''	A5
	byte	$ff			''	A6
	byte	$ff			''	A7
	byte	$ff			''	A8
	byte	$ff			''	A9
	byte	$ff			''	AA
	byte	$ff			''	AB
	byte	$ff			''	AC
	byte	$ff			''	AD
	byte	$ff			''	AE
	byte	$ff			''	AF
	byte	$ff			''	B0
	byte	$ff			''	B1
	byte	$ff			''	B2
	byte	$ff			''	B3
	byte	$ff			''	B4
	byte	$ff			''	B5
	byte	$ff			''	B6
	byte	$ff			''	B7
	byte	$ff			''	B8
	byte	$ff			''	B9
	byte	$ff			''	BA
	byte	$ff			''	BB
	byte	$ff			''	BC
	byte	$ff			''	BD
	byte	$ff			''	BE
	byte	$ff			''	BF

	byte	6<<4|5			''	C0	Left Arrow
	byte	6<<4|6			''	C1	Right Arrow
	byte	6<<4|3			''	C2	Up Arrow
	byte	6<<4|4			''	C3	Down Arrow
	byte	6<<4|1			''	C4	Home
	byte	7<<4|7			''	C5	End
	byte	7<<4|2			''	C6	Page Up
	byte	7<<4|3			''	C7	Page Down
	byte	6<<4|5			''	C8	Backspace
	byte	7<<4|5			''	C9	Delete
	byte	7<<4|4			''	CA	Insert
	byte	6<<4|2			''	CB	Esc
	byte	$ff			''	CC	Apps
	byte	$ff			''	CD	Power
	byte	$ff			''	CE	Sleep
	byte	$ff			''	CF	Wakeup

	byte	$ff			''	D0..DB	F1..F12
	byte	$ff			''	DC	Print Screen
	byte	$ff			''	DD	Scroll Lock
	byte	$ff			''	DE	Caps Lock
	byte	$ff			''	DF	Num Lock

	byte	$ff			''	E0	Keypad 0
	byte	$ff			''	E1	Keypad 1
	byte	$ff			''	E2	Keypad 2
	byte	$ff			''	E3	Keypad 3
	byte	$ff			''	E4	Keypad 4
	byte	$ff			''	E5	Keypad 5
	byte	$ff			''	E6	Keypad 6
	byte	$ff			''	E7	Keypad 7
	byte	$ff			''	E8	Keypad 8
	byte	$ff			''	E8	Keypad 9
	byte	$ff			''	EA	Keypad .
	byte	6<<4|0			''	EB	Keypad Enter
	byte	$ff			''	EC	Keypad +
	byte	$ff			''	ED	Keypad -
	byte	$ff			''	EE	Keypad *
	byte	$ff			''	EF	Keypad /

	byte	7<<4|0			''	F0	Left Shift
	byte	7<<4|0			''	F1	Right Shift
	byte	7<<4|6			''	F2	Left Ctrl
	byte	7<<4|6			''	F3	Right Ctrl
	byte	$ff			''	F4	Left Alt
	byte	$ff			''	F5	Right Alt
	byte	$ff			''	F6	Left Win
	byte	$ff			''	F7	Right Win
	byte	$ff			''	F8
	byte	$ff			''	F9
	byte	$ff			''	FA
	byte	$ff			''	FB
	byte	$ff			''	FC
	byte	$ff			''	FD	Scroll Lock State
	byte	$ff			''	FE	Caps Lock State
	byte	$ff			''	FF	Num Lock State


msg_break	byte	"BC:",0
		byte	" DE:",0
		byte	" HL:",0
		byte	" AF:",0
		byte	" BC'",0
		byte	" DE'",0
		byte	" HL'",0
		byte	" AF'",0
		byte	13,10, "R:",0
		byte	" IFF1:",0
		byte	" IFF2:",0
		byte	" IM:",0
		byte	" I:",0
		byte	" IX:",0
		byte	" IY:",0
		byte	" SP:",0
		byte	" PC:",0
		byte	" = ",0

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
